@@ -5,27 +5,47 @@ import logging
 import queue
 import threading
 from asyncio.events import AbstractEventLoop
-from typing import Callable, Optional, Union
+from typing import Callable, Generic, Optional, TypeVar, Union
 
 from aiortc import RTCPeerConnection, RTCSessionDescription
 from aiortc.contrib.media import MediaPlayer, MediaRecorder
 
-from .receive import VideoReceiver
-from .transform import (
-    AsyncVideoTransformTrack,
+from .process import (
+    AsyncAudioProcessTrack,
+    AsyncVideoProcessTrack,
+    AudioProcessorBase,
+    AudioProcessTrack,
+    VideoProcessorBase,
+    VideoProcessTrack,
     VideoTransformerBase,
-    VideoTransformTrack,
 )
+from .receive import AudioReceiver, VideoReceiver
+
+__all__ = [
+    "AudioProcessorBase",
+    "AudioProcessorFactory",
+    "VideoTransformerBase",
+    "VideoProcessorBase",
+    "MediaPlayerFactory",
+    "MediaRecorderFactory",
+    "VideoProcessorFactory",
+    "WebRtcMode",
+    "TimeoutError",
+    "WebRtcWorker",
+]
+
 
 logger = logging.getLogger(__name__)
 logger.addHandler(logging.NullHandler())
 
 
-VideoTransformFn = Callable
+VideoProcessorT = TypeVar("VideoProcessorT", bound=VideoProcessorBase)
+AudioProcessorT = TypeVar("AudioProcessorT", bound=AudioProcessorBase)
 
 MediaPlayerFactory = Callable[[], MediaPlayer]
 MediaRecorderFactory = Callable[[], MediaRecorder]
-VideoTransformerFactory = Callable[[], VideoTransformerBase]
+VideoProcessorFactory = Callable[[], VideoProcessorT]
+AudioProcessorFactory = Callable[[], AudioProcessorT]
 
 
 class WebRtcMode(enum.Enum):
@@ -45,9 +65,11 @@ async def _process_offer(
     player_factory: Optional[MediaPlayerFactory],
     in_recorder_factory: Optional[MediaRecorderFactory],
     out_recorder_factory: Optional[MediaRecorderFactory],
-    video_transformer: Optional[VideoTransformerBase],
+    video_processor: Optional[VideoProcessorBase],
+    audio_processor: Optional[AudioProcessorBase],
     video_receiver: Optional[VideoReceiver],
-    async_transform: bool,
+    audio_receiver: Optional[AudioReceiver],
+    async_processing: bool,
     callback: Callable[[Union[RTCSessionDescription, Exception]], None],
 ):
     try:
@@ -81,36 +103,51 @@ async def _process_offer(
                     if player and player.audio:
                         logger.info("Add player to audio track")
                         output_track = player.audio
+                    elif audio_processor:
+                        AudioTrack = (
+                            AsyncAudioProcessTrack
+                            if async_processing
+                            else AudioProcessTrack
+                        )
+                        logger.info(
+                            "Add a input audio track %s to "
+                            "output track with audio_processor %s",
+                            input_track,
+                            AudioTrack,
+                        )
+                        output_track = AudioTrack(
+                            track=input_track, processor=audio_processor
+                        )
+                        logger.info("Add the audio track with processor to %s", pc)
                     else:
-                        # Transforming audio is not supported yet.
+                        # Processing audio is not supported yet.
                         output_track = input_track  # passthrough
                 elif input_track.kind == "video":
                     if player and player.video:
                         logger.info("Add player to video track")
                         output_track = player.video
-                    elif video_transformer:
+                    elif video_processor:
                         VideoTrack = (
-                            AsyncVideoTransformTrack
-                            if async_transform
-                            else VideoTransformTrack
+                            AsyncVideoProcessTrack
+                            if async_processing
+                            else VideoProcessTrack
                         )
                         logger.info(
                             "Add a input video track %s to "
-                            "output track with video_transformer %s",
+                            "output track with video_processor %s",
                             input_track,
                             VideoTrack,
                         )
-                        local_video = VideoTrack(
-                            track=input_track, video_transformer=video_transformer
+                        output_track = VideoTrack(
+                            track=input_track, processor=video_processor
                         )
-                        logger.info("Add the video track with transfomer to %s", pc)
-                        output_track = local_video
+                        logger.info("Add the video track with processor to %s", pc)
                     else:
                         output_track = input_track
 
                 if not output_track:
                     raise Exception(
-                        "Neither a player nor a transformer is created. "
+                        "Neither a player nor a processor is created. "
                         "Either factory must be set."
                     )
 
@@ -137,8 +174,11 @@ async def _process_offer(
                 logger.info("Track %s received", input_track.kind)
 
                 if input_track.kind == "audio":
-                    # Not supported yet
-                    pass
+                    if audio_receiver:
+                        logger.info(
+                            "Add a track %s to receiver %s", input_track, audio_receiver
+                        )
+                        audio_receiver.addTrack(input_track)
                 elif input_track.kind == "video":
                     if video_receiver:
                         logger.info(
@@ -155,6 +195,8 @@ async def _process_offer(
                     logger.info("Track %s ended", input_track.kind)
                     if video_receiver:
                         video_receiver.stop()
+                    if audio_receiver:
+                        audio_receiver.stop()
                     if in_recorder:
                         await in_recorder.stop()
 
@@ -179,6 +221,8 @@ async def _process_offer(
 
         if video_receiver and video_receiver.hasTrack():
             video_receiver.start()
+        if audio_receiver and audio_receiver.hasTrack():
+            audio_receiver.start()
 
         if in_recorder:
             await in_recorder.start()
@@ -199,29 +243,46 @@ async def _process_offer(
 webrtc_thread_id_generator = itertools.count()
 
 
-class WebRtcWorker:
+class WebRtcWorker(Generic[VideoProcessorT, AudioProcessorT]):
     _webrtc_thread: Union[threading.Thread, None]
     _loop: Union[AbstractEventLoop, None]
     _answer_queue: queue.Queue
-    _video_transformer: Optional[VideoTransformerBase]
+    _video_processor: Optional[VideoProcessorT]
+    _audio_processor: Optional[AudioProcessorT]
     _video_receiver: Optional[VideoReceiver]
+    _audio_receiver: Optional[AudioReceiver]
 
     @property
-    def video_transformer(self) -> Optional[VideoTransformerBase]:
-        return self._video_transformer
+    def video_processor(self) -> Optional[VideoProcessorT]:
+        return self._video_processor
+
+    @property
+    def audio_processor(self) -> Optional[AudioProcessorT]:
+        return self._audio_processor
 
     @property
     def video_receiver(self) -> Optional[VideoReceiver]:
         return self._video_receiver
 
+    @property
+    def audio_receiver(self) -> Optional[AudioReceiver]:
+        return self._audio_receiver
+
     def __init__(
         self,
         mode: WebRtcMode,
         player_factory: Optional[MediaPlayerFactory] = None,
         in_recorder_factory: Optional[MediaRecorderFactory] = None,
         out_recorder_factory: Optional[MediaRecorderFactory] = None,
-        video_transformer_factory: Optional[VideoTransformerFactory] = None,
-        async_transform: bool = True,
+        video_processor_factory: Optional[
+            VideoProcessorFactory[VideoProcessorT]
+        ] = None,
+        audio_processor_factory: Optional[
+            AudioProcessorFactory[AudioProcessorT]
+        ] = None,
+        async_processing: bool = True,
+        video_receiver_size: int = 4,
+        audio_receiver_size: int = 4,
     ) -> None:
         self._webrtc_thread = None
         self._loop = None
@@ -232,11 +293,16 @@ class WebRtcWorker:
         self.player_factory = player_factory
         self.in_recorder_factory = in_recorder_factory
         self.out_recorder_factory = out_recorder_factory
-        self.video_transformer_factory = video_transformer_factory
-        self.async_transform = async_transform
+        self.video_processor_factory = video_processor_factory
+        self.audio_processor_factory = audio_processor_factory
+        self.async_processing = async_processing
+        self.video_receiver_size = video_receiver_size
+        self.audio_receiver_size = audio_receiver_size
 
-        self._video_transformer = None
+        self._video_processor = None
+        self._audio_processor = None
         self._video_receiver = None
+        self._audio_receiver = None
 
     def _run_webrtc_thread(
         self,
@@ -278,22 +344,30 @@ class WebRtcWorker:
         def callback(localDescription):
             self._answer_queue.put(localDescription)
 
-        video_transformer = None
-        if self.video_transformer_factory:
-            video_transformer = self.video_transformer_factory()
+        video_processor = None
+        if self.video_processor_factory:
+            video_processor = self.video_processor_factory()
+
+        audio_processor = None
+        if self.audio_processor_factory:
+            audio_processor = self.audio_processor_factory()
 
         video_receiver = None
+        audio_receiver = None
         if self.mode == WebRtcMode.SENDONLY:
-            video_receiver = VideoReceiver(queue_maxsize=1)
+            video_receiver = VideoReceiver(queue_maxsize=self.video_receiver_size)
+            audio_receiver = AudioReceiver(queue_maxsize=self.audio_receiver_size)
 
-        self._video_transformer = video_transformer
+        self._video_processor = video_processor
+        self._audio_processor = audio_processor
         self._video_receiver = video_receiver
+        self._audio_receiver = audio_receiver
 
         @self.pc.on("iceconnectionstatechange")
         async def on_iceconnectionstatechange():
             iceConnectionState = self.pc.iceConnectionState
             if iceConnectionState == "closed" or iceConnectionState == "failed":
-                self._unset_transformers()
+                self._unset_processors()
 
         loop.create_task(
             _process_offer(
@@ -303,9 +377,11 @@ class WebRtcWorker:
                 player_factory=self.player_factory,
                 in_recorder_factory=self.in_recorder_factory,
                 out_recorder_factory=self.out_recorder_factory,
-                video_transformer=video_transformer,
+                video_processor=video_processor,
+                audio_processor=audio_processor,
                 video_receiver=video_receiver,
-                async_transform=self.async_transform,
+                audio_receiver=audio_receiver,
+                async_processing=self.async_processing,
                 callback=callback,
             )
         )
@@ -347,12 +423,18 @@ class WebRtcWorker:
 
         return result
 
-    def _unset_transformers(self):
-        self._video_transformer = None
+    def _unset_processors(self):
+        self._video_processor = None
+        self._audio_processor = None
+        if self._video_receiver:
+            self._video_receiver.stop()
         self._video_receiver = None
+        if self._audio_receiver:
+            self._audio_receiver.stop()
+        self._audio_receiver = None
 
     def stop(self, timeout: Union[float, None] = 1.0):
-        self._unset_transformers()
+        self._unset_processors()
         if self._loop:
             self._loop.stop()
         if self._webrtc_thread:
